###############################################################################
#
# MetaSearch2 Catalog Client
#
# Copyright (C) 2024 GBD Consult (info@gbd-consult.de)
#
# This source is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This code is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
###############################################################################
import json
from typing import List, Union
from urllib.parse import urlsplit

from PyQt5.QtCore import QUrl, QUrlQuery
from PyQt5.QtNetwork import QSslCertificate, QSslConfiguration, QNetworkReply, QNetworkRequest
from owslib.catalogue.csw2 import CatalogueServiceWeb, namespaces
import inspect
from io import BytesIO
import re

from owslib.etree import etree
from owslib import util
from owslib import ows
from owslib.util import cleanup_namespaces, bind_url, add_namespaces
from owslib.ogcapi.records import Records
from qgis.core import QgsApplication, QgsNetworkAccessManager

from typing import Optional, Dict, Any
from PyQt5.QtCore import QByteArray
from qgis.core import QgsAuthMethodConfig


def correct_url(url):
    """Sometimes the URL may contain wrong character at the start of the string, this is corrected here

     >>> correct_url("http://test.de")
     'http://test.de'

     >>> correct_url("$http://test.de")
     'http://test.de'

     >>> correct_url("https://test.de")
     'https://test.de'

     >>> correct_url("$https://test.de")
     'https://test.de'

     >>> correct_url("^^^https://test.de")
     'https://test.de'

     >>> correct_url("LOL")
     'LOL'

     """
    # Find the index where 'http', 'https', or 'ftp' starts using regular expression
    match = re.search(r'https?://|ftp://', url)
    if match:
        # Slice the URL from the found index
        return url[match.start():]
    else:
        # If 'http', 'https', or 'ftp' is not found, return the original URL
        return url


def add_params_to_url(qurl: QUrl, params: Optional[Dict[str, Any]]):
    """Add all parameters from a dictionary to a QUrl

    :param qurl: The url
    :param params: The parameter dictionary
    """
    if params:
        query = QUrlQuery()
        for key, value in params.items():
            query.addQueryItem(key, str(value))
        qurl.setQuery(query)


def add_headers_to_request(request: QNetworkRequest, headers: Optional[Dict[str, str]]):
    if headers:
        for key, value in headers.items():
            request.setRawHeader(key.encode(), value.encode())


def add_authentication_to_request(request: QNetworkRequest, qgs_auth: QgsAuthMethodConfig):
    """Check for authentication and add it to the header

    :param request: The HTTP(S) request to add authentication
    :param qgs_auth: The QGIS authentication object
    """
    if qgs_auth and qgs_auth.hasConfig("username") and qgs_auth.hasConfig("password"):
        username = qgs_auth.config("username")
        password = qgs_auth.config("password")
        auth_header_value = QByteArray(f"{username}:{password}".encode()).toBase64()
        request.setRawHeader(b"Authorization", b"Basic " + auth_header_value)


def configure_ssl_for_request(request: QNetworkRequest, qurl: QUrl):
    """Add all trusted SSL certificates to the QUrl object

    :param request: The request
    :param qurl:  The urs
    """
    if qurl.scheme() == "https":
        trusted_cas: List[QSslCertificate] = QgsApplication.authManager().trustedCaCertsCache()
        ssl_config = QSslConfiguration.defaultConfiguration()
        ssl_config.addCaCertificates(trusted_cas)
        request.setSslConfiguration(ssl_config)


def build_network_request(url: str, qgs_auth: QgsAuthMethodConfig,
                          headers: Optional[Dict[str, str]] = None,
                          params: Optional[Dict[str, Any]] = None) -> QNetworkRequest:
    """
    Create a QNetworkRequest including SSL certificates and authentication, headers, and parameters.

    :param url: URL for the request.
    :param qgs_auth: QgsAuthMethodConfig for authentication.
    :param headers: Optional dictionary for request headers.
    :param params: Optional dictionary for URL parameters.
    :return: Configured QNetworkRequest.
    """
    url = correct_url(url)
    print(url)  # To help debugging URL problems, print the URL to the QGIS Python console
    qurl = QUrl(url)
    if not qurl.isValid():
        raise ValueError(f"Invalid URL: {url}")

    add_params_to_url(qurl, params)
    request = QNetworkRequest(qurl)
    add_headers_to_request(request, headers)
    add_authentication_to_request(request, qgs_auth)
    configure_ssl_for_request(request, qurl)

    return request


def network_access_manager_get(url: str, qgs_auth: QgsAuthMethodConfig,
                               headers: Optional[Dict[str, str]] = None,
                               params: Optional[Dict[str, Any]] = None,
                               timeout: int = 10) -> QByteArray:
    """
    Perform a blocking HTTP GET request using the QgsNetworkAccessManager.

    :param url: URL for the request.
    :param qgs_auth: QgsAuthMethodConfig for authentication.
    :param headers: Optional dictionary for request headers.
    :param params: Optional dictionary for URL parameters.
    :param timeout: Timeout for the request in seconds.
    :return: Content of the response as QByteArray.
    :raises ValueError: If an error occurs in the network request.
    """
    if not url:
        raise ValueError("Correct URL is missing.")

    request = build_network_request(url=url, qgs_auth=qgs_auth, headers=headers, params=params)
    manager = QgsNetworkAccessManager.instance()
    manager.setTimeout(timeout * 1000)  # Convert seconds to milliseconds
    reply = manager.blockingGet(request)
    error = reply.error()
    if error == QNetworkReply.NoError:
        return reply.content()
    else:
        error_message = f"URL {url} Error: {reply.errorString()}"
        raise ValueError(error_message)


def network_access_manager_post(url: str, qgs_auth: QgsAuthMethodConfig, data: Any,
                                headers: Optional[Dict[str, str]] = None,
                                params: Optional[Dict[str, Any]] = None,
                                timeout: int = 10) -> QByteArray:
    """
    Perform a blocking HTTP POST request using the QgsNetworkAccessManager.

    :param url: URL for the request.
    :param qgs_auth: QgsAuthMethodConfig for authentication.
    :param data: Data to be sent with the POST request.
    :param headers: Optional dictionary for request headers.
    :param params: Optional dictionary for URL parameters.
    :param timeout: Timeout for the request in seconds.
    :return: Content of the response as QByteArray.
    :raises ValueError: If URL is invalid.
    :raises Exception: If there is an error in the network request.
    """
    if not url:
        raise ValueError("URL required")

    owslib_headers = {
        'User-Agent': 'OWSLib (https://owslib.readthedocs.io)',
        'Content-type': 'text/xml',
        'Accept': 'text/xml,application/xml',
        'Accept-Language': 'en-US',
        'Host': urlsplit(url).netloc,
    }
    if isinstance(data, dict):
        owslib_headers['Content-type'] = 'application/json'
        owslib_headers.pop('Accept', None)
        data = json.dumps(data).encode()

    if headers:
        owslib_headers.update(headers)

    request = build_network_request(url=url, qgs_auth=qgs_auth, headers=owslib_headers, params=params)
    manager = QgsNetworkAccessManager.instance()
    manager.setTimeout(timeout * 1000)  # Convert seconds to milliseconds
    reply = manager.blockingPost(request, data)
    error = reply.error()
    if error == QNetworkReply.NoError:
        return reply.content()
    else:
        raise Exception(f"Error in POST request to {url}: {reply.errorString()}")


class NetRecords(Records):
    """
    A class for managing network records, extending functionality from a base 'Records' class.
    This class handles network operations to retrieve and send data using HTTP GET and POST requests.

    Attributes:
        qgs_auth (QgsAuthMethodConfig): Authentication configuration for network requests.
        json_ (str): JSON formatted string representing the data retrieved from the initial GET request.
    """

    def __init__(self, url: str, timeout: int = 30, qgs_auth: QgsAuthMethodConfig = None):
        """
        Initialize the NetRecords instance with a URL and optional authentication and timeout.

        Args:
            url (str): The URL to fetch data from during initialization.
            timeout (int, optional): Timeout for network requests in seconds. Defaults to 30.
            qgs_auth (QgsAuthMethodConfig, optional): Authentication config for network requests.
        """
        super().__init__(url=url, timeout=timeout, json_=None, auth=None)
        self.qgs_auth = qgs_auth
        try:
            data = network_access_manager_get(url, qgs_auth=self.qgs_auth)
            self.json_ = data.data().decode('utf-8')
        except Exception as e:
            # Handle network or decoding error
            print(f"Error initializing NetRecords: {e}")

    def _request(self, method: str = 'GET', path: str = None,
                 data: str = None, as_dict: bool = True,
                 params: dict = None, kwargs: dict = None) -> Union[str, Dict]:
        """
        Perform a network request using the specified method.

        Args:
            method (str, optional): The HTTP method to use ('GET' or 'POST'). Defaults to 'GET'.
            path (str, optional): The path to append to the base URL for the request. Defaults to None.
            data (str, optional): The data to send with the request (used with 'POST'). Defaults to None.
            as_dict (bool, optional): Whether to return the response as a dictionary. If False, returns raw response. Defaults to True.
            params (dict, optional): Additional parameters to pass with the request. Defaults to None.
            params (dict, optional): Additional arguments to pass with the request. Defaults to None.

        Returns:
            Union[str, Dict]: The response from the server, either as a string or a dictionary based on 'as_dict'.
        """
        if params is None and kwargs:
            params = kwargs
        if params is None:
            params = {}
        url = self._build_url(path)
        self.request = url
        response_data = None

        if method == 'GET':
            response_data = network_access_manager_get(url, qgs_auth=self.qgs_auth, params=params)
        elif method == 'POST':
            response_data = network_access_manager_post(url, qgs_auth=self.qgs_auth, data=data, params=params)

        if response_data:
            return self._process_response(response_data, as_dict)

    def _process_response(self, response_data: QByteArray, as_dict: bool) -> Union[str, Dict]:
        """
        Process the network response data.

        Args:
            response_data (QByteArray): The raw response data from the network request.
            as_dict (bool): Flag to determine if the response should be converted to a dictionary.

        Returns:
            Union[str, Dict]: Processed response data, either as a raw string or a dictionary.
        """
        data = response_data.data()
        if as_dict:
            try:
                if len(data) == 0:
                    return {}
                else:
                    return json.loads(data.decode('utf-8'))
            except json.JSONDecodeError:
                # Handle JSON decode error
                return {}
        else:
            return data


class NetCatalogueServiceWeb(CatalogueServiceWeb):
    """
    A class that extends CatalogueServiceWeb to use QgsNetworkAccessManager for network requests.
    This class handles the communication with a CSW (Catalog Service for the Web) server.
    """

    def __init__(self, url, auth: QgsAuthMethodConfig, lang='en-US', version='2.0.2', timeout=10,
                 skip_caps=False):
        """
        Initializes the NetCatalogueServiceWeb instance with the given parameters.

        Args:
            url (str): The base URL of the CSW service.
            auth (QgsAuthMethodConfig): Authentication configuration for network requests.
            lang (str): Language for the CSW service.
            version (str): CSW service version.
            timeout (int): Timeout for network requests in seconds.
            skip_caps (bool): Whether to skip capabilities document retrieval.
        """
        self.qgs_auth = auth
        super().__init__(url, lang, version, timeout, skip_caps)

    def _invoke(self):
        """
        Performs the network request and processes the response.
        """
        # do HTTP request
        request_url = self.url
        # Get correct URL based on Operation list.

        # If skip_caps=True, then self.operations has not been set, so use
        # default URL.
        if hasattr(self, 'operations'):
            caller = inspect.stack()[1][3]
            if caller == 'getrecords2':
                caller = 'getrecords'
            try:
                op = self.get_operation_by_name(caller)
                if isinstance(self.request, str):  # GET KVP
                    get_verbs = [x for x in op.methods if x.get('type').lower() == 'get']
                    request_url = get_verbs[0].get('url')
                else:
                    post_verbs = [x for x in op.methods if x.get('type').lower() == 'post']
                    if len(post_verbs) > 1:
                        # Filter by constraints.  We must match a PostEncoding of "XML"
                        for pv in post_verbs:
                            for const in pv.get('constraints'):
                                if const.name.lower() == 'postencoding':
                                    values = [v.lower() for v in const.values]
                                    if 'xml' in values:
                                        request_url = pv.get('url')
                                        break
                        else:
                            # Well, just use the first one.
                            request_url = post_verbs[0].get('url')
                    elif len(post_verbs) == 1:
                        request_url = post_verbs[0].get('url')
            except Exception:  # no such luck, just go with request_url
                pass

        if isinstance(self.request, str):  # GET KVP
            self.request = '%s%s' % (bind_url(request_url), self.request)
            self.response = network_access_manager_get(url=self.request, qgs_auth=self.qgs_auth)
        else:
            self.request = cleanup_namespaces(self.request)
            # Add any namespaces used in the "typeNames" attribute of the
            # csw:Query element to the query's xml namespaces.
            for query in self.request.findall(util.nspath_eval('csw:Query', namespaces)):
                ns = query.get("typeNames", None)
                if ns is not None:
                    # Pull out "gmd" from something like "gmd:MD_Metadata" from the list
                    # of typenames
                    ns_keys = [x.split(':')[0] for x in ns.split(' ')]
                    self.request = add_namespaces(self.request, ns_keys)
            self.request = add_namespaces(self.request, 'ows')

            self.request = util.element_to_string(self.request, encoding='utf-8')
            data = network_access_manager_post(url=request_url, data=self.request, qgs_auth=self.qgs_auth,
                                               headers=self.headers, timeout=self.timeout)
            self.response = data.data()
        # parse result see if it's XML
        self._exml = etree.parse(BytesIO(self.response))

        # it's XML.  Attempt to decipher whether the XML response is CSW-ish """
        valid_xpaths = [
            util.nspath_eval('ows:ExceptionReport', namespaces),
            util.nspath_eval('csw:Capabilities', namespaces),
            util.nspath_eval('csw:DescribeRecordResponse', namespaces),
            util.nspath_eval('csw:GetDomainResponse', namespaces),
            util.nspath_eval('csw:GetRecordsResponse', namespaces),
            util.nspath_eval('csw:GetRecordByIdResponse', namespaces),
            util.nspath_eval('csw:HarvestResponse', namespaces),
            util.nspath_eval('csw:TransactionResponse', namespaces)
        ]

        if self._exml.getroot().tag not in valid_xpaths:
            raise RuntimeError('Document is XML, but not CSW-ish')

        # check if it's an OGC Exception
        val = self._exml.find(util.nspath_eval('ows:Exception', namespaces))
        if val is not None:
            raise ows.ExceptionReport(self._exml, self.owscommon.namespace)
        else:
            self.exceptionreport = None
